==> zim_0.15.orig.tar.gz: Zim-0.15/lib/Zim/Formats/Wiki.pm <==
package Zim::Formats::Wiki;

use strict;
no warnings;

our $VERSION = '0.15';

# TODO: some tags can be nested: email links for example
# TODO: header in the middle of a para doesn't get recognized


=head1 NAME

Zim::Formats::Wiki - Wiki text parser

=head1 DESCRIPTION

This is the default parser for Zim.
It uses a wiki-style syntax to format plain text.

FIXME more verbose description

=head1 METHODS

=over 4

=item C<load_tree(IO, PAGE)>

Reads plain text from a filehandle and returns a parse tree.

=cut

sub load_tree { # TODO whitelines between verbatim blocks should be preserved
	my ($class, $io, $page) = @_;

	my @tree;
	my $para = '';
	while (<$io>) {
		if (!/\S/) {
			push @tree, $class->parse_para($para) if $para =~ /\S/;
			$tree[-1][1]{empty_lines}++ if scalar @tree;
			$para = '';
		}
		elsif (/^==+\s+\S+/) {
			push @tree, $class->parse_para($para) if $para =~ /\S/;
			push @tree, $class->parse_head($_);
			$para = '';
		}
		else { $para .= $_ }
	}
	push @tree, $class->parse_para($para) if $para =~ /\S/;
	
	#use Data::Dumper; print STDERR Dumper \@tree;
	return ['Document', {}, @tree];
}

sub parse_head { # parse a header
	my ($class, $head) = @_;
	chomp $head;
	$head =~ s/^(==+)\s+(.*?)(\s+==+|\s*)$/$2/;
	my $level = 7 - length($1); # =X6 => head1, =X5 => head2 etc.
	$level = 1 if $level < 1;
	return ['head'.$level, {}, $head];
}

sub parse_para {
	my ($class, $text) = @_;
	return ( ($text =~ /^(?!\t|\s\s+)/s) # one tab or two spaces
		? ['Para', {}, $class->parse_block($text)]
		: $class->parse_Verbatim($text)            );
}

sub parse_Verbatim { # parse pre formated paragraphs
	my ($class, $pre) = @_;
	my ($indent) = ($pre =~ /^(\s+)/);
	$pre =~ s/^$indent//mg;
	return ['Verbatim', {}, $pre];
}

our @parser_subs = qw/parse_verbatim parse_link parse_styles parse_image/;

sub parse_block { # parse a block of text
	my ($class, @text) = @_;
	for my $sub (@parser_subs) {
		@text = map {ref($_) ? $_ : ($class->$sub($_))} @text;
	}
	return @text;
}

sub parse_verbatim { # like parse_style() but needs to be done earlier
	my ($class, $text) = @_;
	my $i = 0;
	return	map {
		unless ($i++ % 2) { $_ }
		elsif (/^\'\'(.+)\'\'$/) { ['verbatim',  {}, $1] }
	} split /(\'\'.+?\'\')/, $text;
}

sub parse_link { # parse links and urls
	my ($class, $text) = @_;
	my $i = 0;
	return map {
		if ($i++ % 2) {
			my $text;
			if (s/^\[\[(.+)\|(.+)\]\]$/$1/) { $text = $2 }
			else { s/^\[\[(.+)\]\]$/$1/ }
			
			if (/^mailto:|^\S+\@\S+\.\w+$/) {
				$text = $_ unless length $text;
				$_ =~ s/^(mailto:)?/mailto:/;
			}
			
			['link', {to => $_}, length($text) ? $text : $_ ]
		} else { $_ }
	} split /(
		\[\[.+?\]\]                              |
		\b\w+:\/\/     \S*\[\S+\](?:\S+[\w\/])?  |
		\b\w+:\/\/     \S+[\w\/]                 |
		\bmailto:\S+\@ \S*\[\S+\](?:\S+[\w\/])?  |
		\bmailto:\S+\@ \S+[\w\/]                 |
		\b\S+\@\S+\.\w+\b
	)/x, $text;
		# The host name in an uri can be "[hex:hex:..]" for ipv6
		# but we do not want to match "[http://foo.org]"
		# See rfc3986 for the official -but unpractical- regex
}

sub parse_styles { # parse blocks of bold, italic and underline
	my ($class, $text) = @_;
	my $i = 0;
	return	map {
		unless ($i++ % 2) { $_ }
		elsif (/^\*\*(.+)\*\*$/) { ['bold',      {}, $1] }
		elsif (/^\/\/(.+)\/\/$/) { ['italic',    {}, $1] }
		elsif (/^\~\~(.+)\~\~$/) { ['strike',    {}, $1] }
		elsif (/^\_\_(.+)\_\_$/) { ['underline', {}, $1] }
	} split /(
		\*\*.+?\*\* |
		(?<!\:)\/\/.+?\/\/ |
		\~\~.+?\~\~ |
		__.+?__
	)/x, $text;
}

sub parse_image {
	my ($class, $text) = @_;
	my $i = 0;
	return map {
		($i++ % 2) ? ['image', {src => $_}] : $_
	} split /\{\{(.+?)\}\}/, $text;
}

=item C<save_tree(IO, TREE, PAGE)>

Serializes the parse tree into a piece of plain text and writes this
to a filehandle.

=cut

sub save_tree {
	# TODO add support for recursive tags
	my ($class, $io, $tree) = @_;

	my $old_fh = select $io;
	eval { $class->_save_tree($tree) };
	select $old_fh;
	die $@ if $@;
}

sub _save_tree {
	my ($class, $tree) = @_;
	
	my ($name, $opt) = splice @$tree, 0, 2;
	die "Invalid parse tree"
		unless length $name and ref($opt) eq 'HASH';

	while (@$tree) {
		my $node = shift @$tree;
		unless (ref $node) {
			print $node;
			next;
		}
	
		my ($tag, $meta) = @$node[0, 1];
		if ($tag eq 'Para') {
			$class->_save_tree($node);
			print "\n"x$$meta{empty_lines} if $$meta{empty_lines};
			next;
		}
		
		my $text = _dump($node);		
		if ($tag eq 'Verbatim')  {
			$text =~ s/^/\t/mg;
			print $text;
		}
		elsif ($tag =~ /^head(\d)$/) {
			my $n = 7 - $1;
			print ( ('='x$n)." $text ".('='x$n)."\n" );
		}
		elsif ($tag eq 'image') {
			print '{{'.$$meta{src}.'}}';
		}
		elsif ($tag eq 'link') {
			my $to = $$meta{to};
			$to =~ s/^mailto:// unless $text =~ /^mailto:/;
			print (
				($to ne $text) ? "[[$to|$text]]" :
				($to !~ /\s/ and $to =~ m#^\w+://|^(?:mailto:)?\S+\@\S+\.\w+#)
					? $to : "[[$to]]" );
		}
		# per line markup for remaining tags ...
		elsif ($tag eq 'bold') {
			print map { /\S/ ? "**$_**" : $_} split /(\n)/, $text
		}
		elsif ($tag eq 'italic') {
			print map { /\S/ ? "//$_//" : $_} split /(\n)/, $text
		}
		elsif ($tag eq 'strike') {
			print map { /\S/ ? "~~$_~~" : $_} split /(\n)/, $text
		}
		elsif ($tag eq 'underline') {
			print map { /\S/ ? "__$_\__" : $_} split /(\n)/, $text
		}
		elsif ($tag eq 'verbatim') {
			print map { /\S/ ? "''$_''" : $_} split /(\n)/, $text
		}
		else { die "Unkown tag in wiki parse tree: $tag\n" }
		
		print "\n"x$$meta{empty_lines} if $$meta{empty_lines};
	}
}

sub _dump {
	my $node = shift;
	splice @$node, 0, 2;
	for (@$node) { $_ = _dump($_) if ref $_ }
	return join '', @$node;
}

1;

__END__

=back

=head1 AUTHOR

Jaap Karssenberg (Pardus) E<lt>pardus@cpan.orgE<gt>

Copyright (c) 2005 Jaap G Karssenberg. All rights reserved.
This program is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.

=head1 SEE ALSO

L<Zim>,
L<Zim::Repository>,
L<Zim::Page>

=cut

==> zim_0.17.orig.tar.gz: Zim-0.17/lib/Zim/Formats/Wiki.pm <==
package Zim::Formats::Wiki;

use strict;
no warnings;

our $VERSION = '0.17';

# TODO: some tags can be nested: email links for example
# TODO: header in the middle of a para doesn't get recognized


=head1 NAME

Zim::Formats::Wiki - Wiki text parser

=head1 DESCRIPTION

This is the default parser for Zim.
It uses a wiki-style syntax to format plain text.

FIXME more verbose description

=head1 METHODS

=over 4

=item C<load_tree(IO, PAGE)>

Reads plain text from a filehandle and returns a parse tree.

=cut

sub load_tree { # TODO whitelines between verbatim blocks should be preserved
	my ($class, $io, $page) = @_;

	my @tree;
	my $para = '';
	while (<$io>) {
		s/\r?\n$/\n/; # DOS to Unix conversion
		if (!/\S/) {
			push @tree, $class->parse_para($para) if $para =~ /\S/;
			$tree[-1][1]{empty_lines}++ if scalar @tree;
			$para = '';
		}
		elsif (/^==+\s+\S+/) {
			push @tree, $class->parse_para($para) if $para =~ /\S/;
			push @tree, $class->parse_head($_);
			$para = '';
		}
		else { $para .= $_ }
	}
	push @tree, $class->parse_para($para) if $para =~ /\S/;
	
	#use Data::Dumper; print STDERR Dumper \@tree;
	return ['Document', {}, @tree];
}

sub parse_head { # parse a header
	my ($class, $head) = @_;
	chomp $head;
	$head =~ s/^(==+)\s+(.*?)(\s+==+|\s*)$/$2/;
	my $level = 7 - length($1); # =X6 => head1, =X5 => head2 etc.
	$level = 1 if $level < 1;
	return ['head'.$level, {}, $head];
}

sub parse_para { # distinguise normal paras from verbatim paras
	my ($class, $text) = @_;
	return  $class->parse_Verbatim($text) unless $text =~ /^(?!\t|\s\s+)/m;
	#$text =~ s/^\.//mg unless $text =~ /^(?!\.?\t|\.?\s\s+)/s;
	return ['Para', {}, $class->parse_block($text)];
}

sub parse_Verbatim { # parse pre formated paragraphs
	my ($class, $pre) = @_;
	my ($indent) = ($pre =~ /^(\s+)/);
	$pre =~ s/^$indent//mg;
	return ['Verbatim', {}, $pre];
}

our @parser_subs = qw/parse_verbatim parse_links parse_images parse_styles parse_urls/;

sub parse_block { # parse a block of text
	my ($class, @text) = @_;
	for my $sub (@parser_subs) {
		@text = map {ref($_) ? $_ : ($class->$sub($_))} @text;
	}
	return @text;
}

sub parse_verbatim { # like parse_style() but needs to be done earlier
	my ($class, $text) = @_;
	my $i = 0;
	return	map {
		unless ($i++ % 2) { $_ }
		elsif (/^\'\'(.+)\'\'$/) { ['verbatim',  {}, $1] }
	} split /(\'\'.+?\'\')/, $text;
}

sub parse_links {
	my ($class, $text) = @_;
	my $i = 0;
	return
		map   { ($i++ % 2) ? _parse_link($_) : $_ }
		split m#(\[\[.+?\]\])#, $text;
}

sub _parse_link {
	my $l = shift;
	my $text;
	if ($l =~ /^\[\[([^|]+)\|?(.*)\]\]$/) { # [[link]] or [[link|text]]
		($l, $text) = ($1, $2);
	}
	if ($l =~ /^mailto:|^\S+\@\S+\.\w+$/) {
		$text = $l unless length $text;
		$l =~ s/^(mailto:)?/mailto:/;
	}
	return ['link', {to => $l}, length($text) ? $text : $l ];
}

sub parse_urls {
	my ($class, $text) = @_;
	my $i = 0;
	return
		map   { ($i++ % 2) ? _parse_link($_) : $_ }
		split m#(
			\b\w[\w\+\-\.]+://  \S*\[\S+\](?:\S+[\w\/])? |
			\b\w[\w\+\-\.]+://  \S+[\w\/]                |
			\bmailto:\S+\@      \S*\[\S+\](?:\S+[\w/])?  |
			\bmailto:\S+\@      \S+[\w/]                 |
			\b\S+\@\S+\.\w+\b
		)#x, $text;
		# The host name in an uri can be "[hex:hex:..]" for ipv6
		# but we do not want to match "[http://foo.org]"
		# See rfc3986 for the official -but unpractical- regex
}

sub parse_styles { # parse blocks of bold, italic and underline
	my ($class, $text) = @_;
	my $i = 0;
	return	map {
		unless ($i++ % 2) { $_ }
		elsif (/^\*\*(.+)\*\*$/) { ['bold',      {}, $1] }
		elsif (/^\/\/(.+)\/\/$/) { ['italic',    {}, $1] }
		elsif (/^\~\~(.+)\~\~$/) { ['strike',    {}, $1] }
		elsif (/^\_\_(.+)\_\_$/) { ['underline', {}, $1] }
	} split /(
		\*\*.+?\*\* |
		(?<!\:)\/\/.+?\/\/ |
		\~\~.+?\~\~ |
		__.+?__
	)/x, $text;
}

sub parse_images {
	my ($class, $text) = @_;
	my $i = 0;
	return
		map   { ($i++ % 2) ? ['image', {src => $_}] : $_ }
		split /\{\{(.+?)\}\}/, $text;
}

=item C<save_tree(IO, TREE, PAGE)>

Serializes the parse tree into a piece of plain text and writes this
to a filehandle.

=cut

sub save_tree {
	# TODO add support for recursive tags
	my ($class, $io, $tree) = @_;

	my $old_fh = select $io;
	eval { $class->_save_tree($tree) };
	select $old_fh;
	die $@ if $@;
}

sub _save_tree {
	my ($class, $tree) = @_;
	
	my ($name, $opt) = splice @$tree, 0, 2;
	die "Invalid parse tree"
		unless length $name and ref($opt) eq 'HASH';

	while (@$tree) {
		my $node = shift @$tree;
		unless (ref $node) {
			print $node;
			next;
		}
	
		my ($tag, $meta) = @$node[0, 1];
		if ($tag eq 'Para') {
			$class->_save_tree($node);
			print "\n"x$$meta{empty_lines} if $$meta{empty_lines};
			next;
		}
		
		my $text = _dump($node);		
		if ($tag eq 'Verbatim')  {
			$text =~ s/^/\t/mg;
			print $text;
		}
		elsif ($tag =~ /^head(\d)$/) {
			my $n = 7 - $1;
			print ( ('='x$n)." $text ".('='x$n)."\n" );
		}
		elsif ($tag eq 'image') {
			print '{{'.$$meta{src}.'}}';
		}
		elsif ($tag eq 'link') {
			my $to = $$meta{to};
			$to =~ s/^mailto:// unless $text =~ /^mailto:/;
			print (
				($to ne $text) ? "[[$to|$text]]" :
				($to !~ /\s/ and $to =~ m#^\w[\w\+\-\.]+://|^(?:mailto:)?\S+\@\S+\.\w+#)
					? $to : "[[$to]]" );
		}
		# per line markup for remaining tags ...
		elsif ($tag eq 'bold') {
			print map { /\S/ ? "**$_**" : $_} split /(\n)/, $text
		}
		elsif ($tag eq 'italic') {
			print map { /\S/ ? "//$_//" : $_} split /(\n)/, $text
		}
		elsif ($tag eq 'strike') {
			print map { /\S/ ? "~~$_~~" : $_} split /(\n)/, $text
		}
		elsif ($tag eq 'underline') {
			print map { /\S/ ? "__$_\__" : $_} split /(\n)/, $text
		}
		elsif ($tag eq 'verbatim') {
			print map { /\S/ ? "''$_''" : $_} split /(\n)/, $text
		}
		else { die "Unkown tag in wiki parse tree: $tag\n" }
		
		print "\n"x$$meta{empty_lines} if $$meta{empty_lines};
	}
}

sub _dump {
	my $node = shift;
	splice @$node, 0, 2;
	for (@$node) { $_ = _dump($_) if ref $_ }
	return join '', @$node;
}

1;

__END__

=back

=head1 AUTHOR

Jaap Karssenberg (Pardus) E<lt>pardus@cpan.orgE<gt>

Copyright (c) 2005 Jaap G Karssenberg. All rights reserved.
This program is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.

=head1 SEE ALSO

L<Zim>,
L<Zim::Repository>,
L<Zim::Page>

=cut
